/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// models.c -- model loading and caching

// models are the only shared resource between a client and server running
// on the same machine.

#include "quakedef.h"

model_t	*loadmodel;
char	loadname[32];	// for hunk tags

void Mod_LoadSpriteModel (model_t *mod, void *buffer);
void Mod_LoadBrushModel (model_t *mod, void *buffer);
void Mod_LoadAliasModel (model_t *mod, void *buffer);
model_t *Mod_LoadModel (model_t *mod, qboolean crash);

byte	mod_novis[MAX_MAP_LEAFS/8];

#define	MAX_MOD_KNOWN	512
model_t	mod_known[MAX_MOD_KNOWN];
int		mod_numknown;
#define	MAX_LBM_HEIGHT		480

#define ALIAS_BASE_SIZE_RATIO		(1.0 / 11.0)
					// normalizing factor so player model works out to about
					//  1 pixel per triangle

/*
===============
Mod_Init
===============
*/
void Mod_Init (void)
{
	memset (mod_novis, 0xff, sizeof(mod_novis));
}

/*
===============
Mod_Init

Caches the data if needed
===============
*/
void *Mod_Extradata (model_t *mod)
{
	void	*r;
	
	r = Cache_Check (&mod->cache);
	if (r)
		return r;

	Mod_LoadModel (mod, true);
	
	if (!mod->cache.data)
		Sys_Error ("Mod_Extradata: caching failed");
	return mod->cache.data;
}

/*
===============
Mod_PointInLeaf
===============
*/
mleaf_t *Mod_PointInLeaf (vec3_t p, model_t *model)
{
	mnode_t		*node;
	float		d;
	mplane_t	*plane;
	
	if (!model || !model->nodes)
		Sys_Error ("Mod_PointInLeaf: bad model");

	node = model->nodes;
	while (1)
	{
		if (node->contents < 0)
			return (mleaf_t *)node;
		plane = node->plane;
		d = DotProduct (p,plane->normal) - plane->dist;
		if (d > 0)
			node = node->children[0];
		else
			node = node->children[1];
	}
	
	return NULL;	// never reached
}


/*
===================
Mod_DecompressVis
===================
*/
byte *Mod_DecompressVis (byte *in, model_t *model)
{
	static byte	decompressed[MAX_MAP_LEAFS/8];
	int		c;
	byte	*out;
	int		row;

	row = (model->numleafs+7)>>3;	
	out = decompressed;

	if (!in)
	{	// no vis info, so make all visible
		while (row)
		{
			*out++ = 0xff;
			row--;
		}
		return decompressed;		
	}

	do
	{
		if (*in)
		{
			*out++ = *in++;
			continue;
		}
	
		c = in[1];
		in += 2;

		while (c)
		{
			*out++ = 0;
			c--;
		}
	} while (out - decompressed < row);
	
	return decompressed;
}

byte *Mod_LeafPVS (mleaf_t *leaf, model_t *model)
{
	// solid leaf
	if (leaf == model->leafs)
		return mod_novis;

	// no vis data
	if (!leaf->visdata)
		return mod_novis;

	// return decompressed
	return leaf->visdata;
}

/*
===================
Mod_ClearAll
===================
*/
void Mod_ClearAll (void)
{
	int		i;
	model_t	*mod;
	
	for (i=0 , mod=mod_known ; i<mod_numknown ; i++, mod++)
		if (mod->type != mod_alias)
			mod->needload = true;
}

/*
==================
Mod_FindName

==================
*/
model_t *Mod_FindName (char *name)
{
	int		i;
	model_t	*mod;
	
	if (!name[0])
		Sys_Error ("Mod_ForName: NULL name");
		
//
// search the currently loaded models
//
	for (i=0 , mod=mod_known ; i<mod_numknown ; i++, mod++)
		if (!strcmp (mod->name, name) )
			break;
			
	if (i == mod_numknown)
	{
		if (mod_numknown == MAX_MOD_KNOWN)
			Sys_Error ("mod_numknown == MAX_MOD_KNOWN");
		strcpy (mod->name, name);
		mod->needload = true;
		mod_numknown++;
	}

	return mod;
}

/*
==================
Mod_TouchModel

==================
*/
void Mod_TouchModel (char *name)
{
	model_t	*mod;
	
	mod = Mod_FindName (name);
	
	if (!mod->needload)
	{
		if (mod->type == mod_alias)
			Cache_Check (&mod->cache);
	}
}

/*
==================
Mod_LoadModel

Loads a model into the cache
==================
*/
model_t *Mod_LoadModel (model_t *mod, qboolean crash)
{
	void	*d;
	unsigned *buf;
	byte	stackbuf[1024];

	if (!mod->needload)
	{
		if (mod->type == mod_alias)
		{
			d = Cache_Check (&mod->cache);

			if (d) return mod;
		}
		else return mod;
	}

	// load the file
	buf = (unsigned *) COM_LoadStackFile (mod->name, stackbuf, sizeof (stackbuf));

	if (!buf)
	{
		if (crash) Sys_Error ("Mod_NumForName: %s not found", mod->name);
		return NULL;
	}

	// allocate a new model
	COM_FileBase (mod->name, loadname);

	loadmodel = mod;

	// call the apropriate loader
	mod->needload = false;

	switch (LittleLong (*(unsigned *) buf))
	{
	case IDPOLYHEADER:
		Mod_LoadAliasModel (mod, buf);
		break;

	case IDSPRITEHEADER:
		Mod_LoadSpriteModel (mod, buf);
		break;

	default:
		Mod_LoadBrushModel (mod, buf);
		break;
	}

	return mod;
}

/*
==================
Mod_ForName

Loads in a model for the given name
==================
*/
model_t *Mod_ForName (char *name, qboolean crash)
{
	model_t	*mod;
	
	mod = Mod_FindName (name);
	
	return Mod_LoadModel (mod, crash);
}


/*
===============================================================================

					BRUSHMODEL LOADING

===============================================================================
*/

byte	*mod_base;


/*
=================
Mod_LoadTextures
=================
*/
void Mod_LoadTextures (lump_t *l)
{
	int		i, j, pixels, num, max, altmax;
	miptex_t	*mt;
	texture_t	*tx, *tx2;
	texture_t	*anims[10];
	texture_t	*altanims[10];
	dmiptexlump_t *m;

	if (!l->filelen)
	{
		loadmodel->textures = NULL;
		return;
	}

	m = (dmiptexlump_t *) (mod_base + l->fileofs);

	m->nummiptex = LittleLong (m->nummiptex);

	loadmodel->numtextures = m->nummiptex;
	loadmodel->textures = Hunk_AllocName (m->nummiptex * sizeof(*loadmodel->textures), loadname);

	for (i = 0; i < m->nummiptex; i++)
	{
		m->dataofs[i] = LittleLong (m->dataofs[i]);

		if (m->dataofs[i] == -1) continue;

		mt = (miptex_t *) ((byte *) m + m->dataofs[i]);

		mt->width = LittleLong (mt->width);
		mt->height = LittleLong (mt->height);

		if ((mt->width & 15) || (mt->height & 15))
			Sys_Error ("Texture %s is not 16 aligned", mt->name);

		pixels = mt->width * mt->height;
		tx = Hunk_AllocName (sizeof (texture_t) + pixels, loadname);
		loadmodel->textures[i] = tx;

		Q_memcpy (tx->name, mt->name, sizeof (tx->name));

		tx->width = mt->width;
		tx->height = mt->height;

		// the pixels immediately follow the structures
		Q_memcpy (tx + 1, mt + 1, pixels);

		// no texturechain
		tx->texturechain = NULL;

		if (!Q_strncmp (mt->name, "sky", 3))
		{
			R_InitSky (tx);
		}
		else
		{
			int upload_flags;

			// set different flags for certain texture types
			if (mt->name[0] == '*')
				upload_flags = (TEX_LIQUID | TEX_BSP | TEX_MIPMAP);
			else upload_flags = (TEX_MIPMAP | TEX_BSP | TEX_LUMASPEC);

			tx->glt = GL_LoadTexture (mt->name, tx->width, tx->height, (byte *) (tx + 1), upload_flags);
		}
	}

	// sequence the animations
	for (i = 0; i < m->nummiptex; i++)
	{
		tx = loadmodel->textures[i];

		// does not exist or not animating
		if (!tx || tx->name[0] != '+') continue;

		// already sequenced
		if (tx->anim_next) continue;

		// find the number of frames in the animation
		memset (anims, 0, sizeof (anims));
		memset (altanims, 0, sizeof (altanims));

		max = tx->name[1];
		altmax = 0;

		if (max >= 'a' && max <= 'z') max -= 'a' - 'A';

		if (max >= '0' && max <= '9')
		{
			max -= '0';
			altmax = 0;
			anims[max] = tx;
			max++;
		}
		else if (max >= 'A' && max <= 'J')
		{
			altmax = max - 'A';
			max = 0;
			altanims[altmax] = tx;
			altmax++;
		}
		else Sys_Error ("Bad animating texture %s", tx->name);

		for (j = i + 1; j < m->nummiptex; j++)
		{
			tx2 = loadmodel->textures[j];

			if (!tx2 || tx2->name[0] != '+') continue;
			if (strcmp (tx2->name + 2, tx->name + 2)) continue;

			num = tx2->name[1];

			if (num >= 'a' && num <= 'z') num -= 'a' - 'A';

			if (num >= '0' && num <= '9')
			{
				num -= '0';
				anims[num] = tx2;

				if (num + 1 > max) max = num + 1;
			}
			else if (num >= 'A' && num <= 'J')
			{
				num = num - 'A';
				altanims[num] = tx2;

				if (num + 1 > altmax) altmax = num + 1;
			}
			else Sys_Error ("Bad animating texture %s", tx->name);
		}

#define	ANIM_CYCLE	2
		// link them all together
		for (j = 0; j < max; j++)
		{
			tx2 = anims[j];

			if (!tx2) Sys_Error ("Missing frame %i of %s", j, tx->name);

			tx2->anim_total = max * ANIM_CYCLE;
			tx2->anim_min = j * ANIM_CYCLE;
			tx2->anim_max = (j + 1) * ANIM_CYCLE;
			tx2->anim_next = anims[(j + 1) % max];

			if (altmax) tx2->alternate_anims = altanims[0];
		}

		for (j = 0; j < altmax; j++)
		{
			tx2 = altanims[j];

			if (!tx2) Sys_Error ("Missing frame %i of %s", j, tx->name);

			tx2->anim_total = altmax * ANIM_CYCLE;
			tx2->anim_min = j * ANIM_CYCLE;
			tx2->anim_max = (j + 1) * ANIM_CYCLE;
			tx2->anim_next = altanims[(j + 1) % altmax];

			if (max) tx2->alternate_anims = anims[0];
		}
	}
}

/*
=================
Mod_LoadLighting
=================
*/
void Mod_LoadLighting (lump_t *l)
{
	int i;
	int j;
	int rescale;
	byte *light_base;
	char litName[MAX_OSPATH];
	char litFolder[MAX_OSPATH];
	FILE *f;
	litheader_t litHeader;
	int filelen;
	byte *lit_data;
	int gs;

	if (!l->filelen)
	{
		loadmodel->lightdata = NULL;
		return;
	}

	// allocate memory
	loadmodel->lightdata = Hunk_AllocName (l->filelen * 3, loadname);	

	// lightdata pointers
	// we need these for lits too, so do them here
	light_base = (mod_base + l->fileofs);
	lit_data = loadmodel->lightdata;

	// look for a lit file.  first off, make the name
	strcpy (litName, loadmodel->name);

	// work backwards so we're sure that we have the extension
	for (i = strlen (litName) - 1; ; i--)
	{
		// we always expect to find a . here (as in ".bsp")
		if (litName[i] == '.')
		{
			break;
		}
	}

	// add the "lit" extension following the '.'
	litName[i + 1] = 'l';
	litName[i + 2] = 'i';
	litName[i + 3] = 't';

	// now we attempt to open it.  first up, try it on the game folder
	sprintf (litFolder, "%s\\%s", com_gamedir, litName);

	f = fopen (litFolder, "rb");

	if (!f)
	{
		// fall back on id1
		sprintf (litFolder, "id1\\%s", litName);
		f = fopen (litFolder, "rb");

		// no joy there
		if (!f) goto LoadBSPLight;
	}

	// this is a bit tricky now, cos we could have a "start.bsp" in a game folder which is different to the "start.bsp"
	// in id1, so the fallback will break the lighting there.  however, it could be a modified version of the classic
	// start map (say, with a vispatch) or it could be that we don't have a "start.bsp" in the game folder and we want
	// to fall back on id1.  basically, we have to ensure that the lit file correctly matches the map it's associated with.
	// the trick is to ensure that the length of the file (- 8 bytes for the header) equals 3 * the size of the BSP light
	// data.  it's wise to do this test anyway, just to ensure that the reads and the pointers will be valid later on.
	fseek (f, 0, SEEK_END);
	filelen = ftell (f);
	fseek (f, 0, SEEK_SET);

	if (ferror (f)) goto InvalidLit;

	if (filelen - sizeof (litheader_t) != l->filelen * 3)
	{
		// lit file mismatches so load from the BSP
		Con_DPrintf ("Mismatched lit file for %s: %i (LIT) vs %i (BSP)\n", loadmodel->name, filelen - sizeof (litheader_t), l->filelen * 3);
		goto InvalidLit;
	}

	// now we read and validate the header
	fread (&litHeader, sizeof (litheader_t), 1, f);

	if (ferror (f)) goto InvalidLit;

	// show-off!!!
	if (litHeader.version != 1) goto InvalidLit;
	if (*((int *) litHeader.ident) != LIT_IDENT) goto InvalidLit;

	Con_DPrintf ("Opened %s OK: %i\n", litFolder, *((int *) litHeader.ident));

	// it's valid now so read the data
	fread (loadmodel->lightdata, l->filelen * 3, 1, f);

	// done with the file
	fclose (f);

	// now this gets really complicated.  the version of MHColour used to make the most commonly available lit files
	// was more about brute force than finesse, and forcibly squashed the light data into the same space as the original
	// light data was in.  for the most part this worked OK.  however, one or two maps have spots where things didn't
	// quite match up, and nobody seemed to have noticed until the lits were already in wide circulation, by which time
	// it was too late.  so here we get ugly and manually hack the data...
	for (i = 0; i < l->filelen; i++, lit_data += 3)
	{
		// take a greyscale version
		gs = ((int) lit_data[0] * 30 + (int) lit_data[1] * 59 + (int) lit_data[2] * 11) / 100;

		if (!gs)
		{
			// copy from BSP
			// would vertex lighting with glColor3f taking surfs that share the verts work better here?
			lit_data[0] = light_base[i];
			lit_data[1] = light_base[i];
			lit_data[2] = light_base[i];
		}
		else if (gs != light_base[i])
		{
			for (j = 0; j < 3; j++)
			{
				// rescale the lit to match the original light intensity
				rescale = ((int) lit_data[j] * (int) light_base[i]) / gs;

				// clamp
				if (rescale > 255) rescale = 255;

				// copy back
				lit_data[j] = rescale;
			}
		}
	}

	// job done!!!
	return;

InvalidLit:
	// just close the file
	fclose (f);

LoadBSPLight:;
	// expand to 3 component
	for (i = 0; i < l->filelen; i++, lit_data += 3)
	{
		lit_data[0] = light_base[i];
		lit_data[1] = light_base[i];
		lit_data[2] = light_base[i];
	}
}


/*
=================
Mod_LoadEntities
=================
*/
void Mod_LoadEntities (lump_t *l)
{
	if (!l->filelen)
	{
		loadmodel->entities = NULL;
		return;
	}

	loadmodel->entities = Hunk_AllocName ( l->filelen, loadname);	
	Q_memcpy (loadmodel->entities, mod_base + l->fileofs, l->filelen);
}


/*
=================
Mod_LoadVertexes
=================
*/
void Mod_LoadVertexes (lump_t *l)
{
	dvertex_t	*in;
	mvertex_t	*out;
	int			i, count;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->vertexes = out;
	loadmodel->numvertexes = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		out->position[0] = LittleFloat (in->point[0]);
		out->position[1] = LittleFloat (in->point[1]);
		out->position[2] = LittleFloat (in->point[2]);
	}
}

/*
=================
Mod_LoadSubmodels
=================
*/
void Mod_LoadSubmodels (lump_t *l)
{
	dmodel_t	*in;
	dmodel_t	*out;
	int			i, j, count;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->submodels = out;
	loadmodel->numsubmodels = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		for (j=0 ; j<3 ; j++)
		{	// spread the mins / maxs by a pixel
			out->mins[j] = LittleFloat (in->mins[j]) - 1;
			out->maxs[j] = LittleFloat (in->maxs[j]) + 1;
			out->origin[j] = LittleFloat (in->origin[j]);
		}
		for (j=0 ; j<MAX_MAP_HULLS ; j++)
			out->headnode[j] = LittleLong (in->headnode[j]);
		out->visleafs = LittleLong (in->visleafs);
		out->firstface = LittleLong (in->firstface);
		out->numfaces = LittleLong (in->numfaces);
	}
}

/*
=================
Mod_LoadEdges
=================
*/
void Mod_LoadEdges (lump_t *l)
{
	dedge_t *in;
	medge_t *out;
	int 	i, count;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( (count + 1) * sizeof(*out), loadname);	

	loadmodel->edges = out;
	loadmodel->numedges = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		out->v[0] = (unsigned short)LittleShort(in->v[0]);
		out->v[1] = (unsigned short)LittleShort(in->v[1]);
	}
}

/*
=================
Mod_LoadTexinfo
=================
*/
void Mod_LoadTexinfo (lump_t *l)
{
	texinfo_t *in;
	mtexinfo_t *out;
	int 	i, j, count;
	int		miptex;
	float	len1, len2;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->texinfo = out;
	loadmodel->numtexinfo = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		for (j=0 ; j<8 ; j++)
			out->vecs[0][j] = LittleFloat (in->vecs[0][j]);
		len1 = Length (out->vecs[0]);
		len2 = Length (out->vecs[1]);
		len1 = (len1 + len2)/2;
		if (len1 < 0.32)
			out->mipadjust = 4;
		else if (len1 < 0.49)
			out->mipadjust = 3;
		else if (len1 < 0.99)
			out->mipadjust = 2;
		else
			out->mipadjust = 1;
#if 0
		if (len1 + len2 < 0.001)
			out->mipadjust = 1;		// don't crash
		else
			out->mipadjust = 1 / floor( (len1+len2)/2 + 0.1 );
#endif

		miptex = LittleLong (in->miptex);
		out->flags = LittleLong (in->flags);
	
		if (!loadmodel->textures)
		{
			out->texture = &r_notexture_mip;	// checkerboard texture
			out->flags = 0;
		}
		else
		{
			if (miptex >= loadmodel->numtextures)
				Sys_Error ("miptex >= loadmodel->numtextures");

			out->texture = loadmodel->textures[miptex];

			if (!out->texture)
			{
				out->texture = &r_notexture_mip; // texture not found
				out->flags = 0;
			}
		}
	}
}

/*
================
CalcSurfaceExtents

Fills in s->texturemins[] and s->extents[]
================
*/
void CalcSurfaceExtents (msurface_t *s)
{
	float	mins[2], maxs[2], val;
	int		i,j, e;
	mvertex_t	*v;
	mtexinfo_t	*tex;
	int		bmins[2], bmaxs[2];

	mins[0] = mins[1] = 999999;
	maxs[0] = maxs[1] = -99999;

	tex = s->texinfo;
	
	for (i=0 ; i<s->numedges ; i++)
	{
		e = loadmodel->surfedges[s->firstedge+i];
		if (e >= 0)
			v = &loadmodel->vertexes[loadmodel->edges[e].v[0]];
		else
			v = &loadmodel->vertexes[loadmodel->edges[-e].v[1]];
		
		for (j=0 ; j<2 ; j++)
		{
			val = v->position[0] * tex->vecs[j][0] + 
				v->position[1] * tex->vecs[j][1] +
				v->position[2] * tex->vecs[j][2] +
				tex->vecs[j][3];
			if (val < mins[j])
				mins[j] = val;
			if (val > maxs[j])
				maxs[j] = val;
		}
	}

	for (i=0 ; i<2 ; i++)
	{	
		bmins[i] = floor(mins[i]/16);
		bmaxs[i] = ceil(maxs[i]/16);

		s->texturemins[i] = bmins[i] * 16;
		s->extents[i] = (bmaxs[i] - bmins[i]) * 16;
		if ( !(tex->flags & TEX_SPECIAL) && s->extents[i] > 2000) //512 /* 256 */ )
			Sys_Error ("Bad surface extents");
	}
}


/*
=================
Mod_LoadFaces
=================
*/
void Mod_LoadFaces (lump_t *l)
{
	dface_t		*in;
	msurface_t 	*surf;
	int			i, count, surfnum;
	int			planenum, side;

	in = (void *) (mod_base + l->fileofs);

	if (l->filelen % sizeof (*in)) Sys_Error ("MOD_LoadBmodel: funny lump size in %s", loadmodel->name);

	count = l->filelen / sizeof (*in);
	surf = Hunk_AllocName (count * sizeof (*surf), loadname);	

	loadmodel->surfaces = surf;
	loadmodel->numsurfaces = count;

	for (surfnum = 0; surfnum < count; surfnum++, in++, surf++)
	{
		surf->firstedge = LittleLong (in->firstedge);
		surf->numedges = LittleShort (in->numedges);		
		surf->flags = 0;

		planenum = LittleShort (in->planenum);
		side = LittleShort (in->side);

		if (side) surf->flags |= SURF_PLANEBACK;			

		surf->plane = loadmodel->planes + planenum;

		surf->texinfo = loadmodel->texinfo + LittleShort (in->texinfo);

		CalcSurfaceExtents (surf);

		// lighting info
		for (i = 0; i < MAXLIGHTMAPS; i++)
			surf->styles[i] = in->styles[i];

		i = LittleLong (in->lightofs);

		if (i == -1)
			surf->samples = NULL;
		else
			surf->samples = loadmodel->lightdata + (i * 3);

		// no texturechain
		surf->texturechain = NULL;

		// set the drawing flags
		if (!Q_strncmp (surf->texinfo->texture->name, "sky", 3))	// sky
		{
			surf->flags |= SURF_DRAWSKY;

			//GL_SubdivideSurface (surf);	// cut up polygon for warps
			continue;
		}

		// precalculate the surface warp indexes
		// this is valid for liquid and for normal surfs (needed for caustics)
		// we really should be setting a plane coefficient for either the texture matrix or some
		// texgen stuff based on surf->truenormal here, but bleagh.  maybe someday...
		switch (surf->plane->type)
		{
		case PLANE_Y:
			// vertical in line with the world grid
			surf->warpindex[0] = 0;
			surf->warpindex[1] = 2;
			break;

		case PLANE_X:
			// vertical in line with the world grid
			surf->warpindex[0] = 2;
			surf->warpindex[1] = 1;
			break;

		case PLANE_ANYX:
		case PLANE_ANYY:
			// vertical at an angle to the world grid
			surf->warpindex[0] = 2;
			surf->warpindex[1] = 0;
			break;

		case PLANE_Z:
			// flat
		default:
			// same as flat
			surf->warpindex[0] = 0;
			surf->warpindex[1] = 1;
			break;
		}

		if (!Q_strncmp (surf->texinfo->texture->name, "*", 1))		// turbulent
		{
			surf->flags |= SURF_DRAWTURB;

			// identify the type of liquid
			if (strlen (surf->texinfo->texture->name) > 5)
			{
				if (!strnicmp (surf->texinfo->texture->name, "*lava", 5)) surf->flags |= SURF_DRAWLAVA;
				if (!strnicmp (surf->texinfo->texture->name, "*slim", 5)) surf->flags |= SURF_DRAWSLIME;
				if (!strnicmp (surf->texinfo->texture->name, "*tele", 5)) surf->flags |= SURF_DRAWTELE;
			}

			surf->extents[0] = 16384;
			surf->extents[1] = 16384;

			surf->texturemins[0] = -8192;
			surf->texturemins[1] = -8192;

			//GL_SubdivideSurface (surf);	// cut up polygon for warps
			continue;
		}
	}
}


/*
=================
Mod_SetParent
=================
*/
void Mod_SetParent (mnode_t *node, mnode_t *parent)
{
	node->parent = parent;
	if (node->contents < 0)
		return;
	Mod_SetParent (node->children[0], node);
	Mod_SetParent (node->children[1], node);
}

/*
=================
Mod_LoadNodes
=================
*/
void Mod_LoadNodes (lump_t *l)
{
	int			i, j, count, p;
	dnode_t		*in;
	mnode_t 	*out;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->nodes = out;
	loadmodel->numnodes = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		for (j=0 ; j<3 ; j++)
		{
			out->minmaxs[j] = LittleShort (in->mins[j]);
			out->minmaxs[3+j] = LittleShort (in->maxs[j]);
		}
	
		p = LittleLong(in->planenum);
		out->plane = loadmodel->planes + p;

		out->firstsurface = LittleShort (in->firstface);
		out->numsurfaces = LittleShort (in->numfaces);
		
		for (j=0 ; j<2 ; j++)
		{
			p = LittleShort (in->children[j]);
			if (p >= 0)
				out->children[j] = loadmodel->nodes + p;
			else
				out->children[j] = (mnode_t *)(loadmodel->leafs + (-1 - p));
		}
	}
	
	Mod_SetParent (loadmodel->nodes, NULL);	// sets nodes and leafs
}

/*
=================
Mod_LoadVisibilityAndLeafs

Also loads visibility
=================
*/
void Mod_LoadVisibilityAndLeafs (lump_t *l, lump_t *v)
{
	dleaf_t 	*in;
	mleaf_t 	*out;
	int			i, j, count, p;
	byte		*vis;
	int			visptr;

	in = (void *) (mod_base + l->fileofs);

	if (l->filelen % sizeof (*in)) Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);

	count = l->filelen / sizeof (*in);

	// allocate vis first cos leafs and nodes should be in contiguous RAM so that the last leaf + 1 maps to the first node
	// when we do our pointer cast.
	// take 1 extra byte to allow for numbers of leafs that don't divide evenly by 8.
	// we don;t bother checking for un-vised maps.  This will be wasteful of memory in that event, but an un-vised map
	// will always be suboptimal to begin with so it's no big deal.
	loadmodel->visdata = Hunk_AllocName ((count + 1) * (count + 1), loadname);
	visptr = 0;

	// now allocate leafs
	out = Hunk_AllocName (count * sizeof (*out), loadname);	

	loadmodel->leafs = out;
	loadmodel->numleafs = count;

	for (i = 0; i < count; i++, in++, out++)
	{
		for (j = 0; j < 3; j++)
		{
			out->minmaxs[j] = LittleShort (in->mins[j]);
			out->minmaxs[3 + j] = LittleShort (in->maxs[j]);
		}

		p = LittleLong (in->contents);
		out->contents = p;

		out->firstmarksurface = loadmodel->marksurfaces + LittleShort(in->firstmarksurface);
		out->nummarksurfaces = LittleShort(in->nummarksurfaces);
		
		p = LittleLong (in->visofs);

		if (p == -1)
			out->visdata = NULL;
		else
		{
			// set up decompressed visibility
			out->visdata = &loadmodel->visdata[visptr];
			visptr += (count + 1);

			// decompress and copy across.
			// a typical ID1 map will require about 2.5MB for decompressed visdata, compared to < 40K for compressed
			// it's an ouch, but RAM isn't really at a premium these days so it's acceptable enough.
			vis = Mod_DecompressVis (mod_base + v->fileofs + p, loadmodel);
			Q_memcpy (out->visdata, vis, count + 1);
		}

		out->efrags = NULL;

		for (j = 0; j < 4; j++)
			out->ambient_sound_level[j] = in->ambient_level[j];

		// gl underwater warp
		// changed this to check for liquid content types.
		if (out->contents == CONTENTS_WATER || out->contents == CONTENTS_SLIME || out->contents == CONTENTS_LAVA)
		{
			for (j = 0; j < out->nummarksurfaces; j++)
				out->firstmarksurface[j]->flags |= SURF_UNDERWATER;
		}
	}	
}


/*
=================
Mod_LoadClipnodes
=================
*/
void Mod_LoadClipnodes (lump_t *l)
{
	dclipnode_t *in, *out;
	int			i, count;
	hull_t		*hull;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->clipnodes = out;
	loadmodel->numclipnodes = count;

	hull = &loadmodel->hulls[1];
	hull->clipnodes = out;
	hull->firstclipnode = 0;
	hull->lastclipnode = count-1;
	hull->planes = loadmodel->planes;
	hull->clip_mins[0] = -16;
	hull->clip_mins[1] = -16;
	hull->clip_mins[2] = -24;
	hull->clip_maxs[0] = 16;
	hull->clip_maxs[1] = 16;
	hull->clip_maxs[2] = 32;

	hull = &loadmodel->hulls[2];
	hull->clipnodes = out;
	hull->firstclipnode = 0;
	hull->lastclipnode = count-1;
	hull->planes = loadmodel->planes;
	hull->clip_mins[0] = -32;
	hull->clip_mins[1] = -32;
	hull->clip_mins[2] = -24;
	hull->clip_maxs[0] = 32;
	hull->clip_maxs[1] = 32;
	hull->clip_maxs[2] = 64;

	for (i=0 ; i<count ; i++, out++, in++)
	{
		out->planenum = LittleLong(in->planenum);
		out->children[0] = LittleShort(in->children[0]);
		out->children[1] = LittleShort(in->children[1]);
	}
}

/*
=================
Mod_MakeHull0

Deplicate the drawing hull structure as a clipping hull
=================
*/
void Mod_MakeHull0 (void)
{
	mnode_t		*in, *child;
	dclipnode_t *out;
	int			i, j, count;
	hull_t		*hull;
	
	hull = &loadmodel->hulls[0];	
	
	in = loadmodel->nodes;
	count = loadmodel->numnodes;
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	hull->clipnodes = out;
	hull->firstclipnode = 0;
	hull->lastclipnode = count-1;
	hull->planes = loadmodel->planes;

	for (i=0 ; i<count ; i++, out++, in++)
	{
		out->planenum = in->plane - loadmodel->planes;
		for (j=0 ; j<2 ; j++)
		{
			child = in->children[j];
			if (child->contents < 0)
				out->children[j] = child->contents;
			else
				out->children[j] = child - loadmodel->nodes;
		}
	}
}

/*
=================
Mod_LoadMarksurfaces
=================
*/
void Mod_LoadMarksurfaces (lump_t *l)
{	
	int		i, j, count;
	short		*in;
	msurface_t **out;
	
	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->marksurfaces = out;
	loadmodel->nummarksurfaces = count;

	for ( i=0 ; i<count ; i++)
	{
		j = LittleShort(in[i]);
		if (j >= loadmodel->numsurfaces)
			Sys_Error ("Mod_ParseMarksurfaces: bad surface number");
		out[i] = loadmodel->surfaces + j;
	}
}

/*
=================
Mod_LoadSurfedges
=================
*/
void Mod_LoadSurfedges (lump_t *l)
{	
	int		i, count;
	int		*in, *out;
	
	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->surfedges = out;
	loadmodel->numsurfedges = count;

	for ( i=0 ; i<count ; i++)
		out[i] = LittleLong (in[i]);
}


/*
=================
Mod_LoadPlanes
=================
*/
void Mod_LoadPlanes (lump_t *l)
{
	int			i, j;
	mplane_t	*out;
	dplane_t 	*in;
	int			count;
	int			bits;
	
	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*2*sizeof(*out), loadname);	
	
	loadmodel->planes = out;
	loadmodel->numplanes = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		bits = 0;
		for (j=0 ; j<3 ; j++)
		{
			out->normal[j] = LittleFloat (in->normal[j]);
			if (out->normal[j] < 0)
				bits |= 1<<j;
		}

		out->dist = LittleFloat (in->dist);
		out->type = LittleLong (in->type);
		out->signbits = bits;
	}
}

/*
=================
RadiusFromBounds
=================
*/
float RadiusFromBounds (vec3_t mins, vec3_t maxs)
{
	int		i;
	vec3_t	corner;

	for (i=0 ; i<3 ; i++)
	{
		corner[i] = fabs(mins[i]) > fabs(maxs[i]) ? fabs(mins[i]) : fabs(maxs[i]);
	}

	return Length (corner);
}


/*
======================================================================================================

		TRANSLUCENT WATER DETECTION

======================================================================================================
*/

/*
====================
Mod_DetectTransWater

Detects translucent water visibility in a map.  First things first, we check to see if the map has any
liquid in it - if not, we don't even bother and just return false.

If it does, we look for empty leafs that contain water surfs.  When we find one, we get it's visibility
and check that to see if it contains any underwater leafs (detection of one is sufficient) - if so, we 
increment a counter.  If this counter is exactly equal to the number of empty leafs with water surfs, 
then we have a map that has translucent water visibility in it.

It's a pretty exhaustive test that can increase the load times on some particularly large maps, but I've
found that it's the only way to be absolutely sure.

The only other code that this is reliant on is running off decompressed vis, although it can be easily
modified to not rely on that.

Tested on a vispatched ID, an un-vispatched Hipnotic and an un-vispatched Rogue, worked perfectly every
time.  I can't guarantee the same with every single map there is, but it's a pretty damn good base.
====================
*/
qboolean Mod_DetectTransWater (model_t *mod)
{
	int i;
	int j;
	int c;
	mleaf_t *leaf;
	mleaf_t *vleaf;
	msurface_t **mark;
	int watersurfleafs = 0;
	int watertransleafs = 0;

	for (i = 0; i < mod->numtextures; i++)
	{
		// not a texture
		if (!mod->textures[i])
			continue;

		// ignore teleports
		if (!strnicmp (mod->textures[i]->name, "*tele", 5))
			continue;

		// water texture
		if (mod->textures[i]->name[0] == '*') break;
	}

	// no water
	if (i == mod->numtextures)
	{
		Con_DPrintf ("Map %s has no water\n", mod->name);
		return false;
	}

	for (i = 0, leaf = mod->leafs; i < mod->numleafs; i++, leaf++)
	{
		// wrong contents
		if (leaf->contents != CONTENTS_EMPTY)
			continue;

		// no vis data
		if (!leaf->visdata)
			continue;

		// check surfs
		mark = leaf->firstmarksurface;
		c = leaf->nummarksurfaces;

		// look for water
		if (c)
		{
			do
			{
				// break if it's liquid but not teleport
				if ((*mark)->texinfo->texture->name[0] == '*' && !((*mark)->flags & SURF_DRAWTELE))
					break;

				mark++;
			} while (--c);
		}

		// we're only interested in leafs that have liquid surfs
		if (!c) continue;

		// increment number of water surf leafs
		watersurfleafs++;

		// check visibility
		for (j = 0; j < mod->numleafs; j++)
		{
			if (leaf->visdata[j >> 3] & (1 << (j & 7)))
			{
				vleaf = &mod->leafs[j + 1];

				// is it a water leaf?
				// note - we don't ignore teleports here cos they are valid for a translucent water vis.
				if (vleaf->contents == CONTENTS_WATER || vleaf->contents == CONTENTS_SLIME || vleaf->contents == CONTENTS_LAVA)
				{
					// increment counter and break, cos we need a 1 for 1 match in the counters
					watertransleafs++;
					break;
				}
			}
		}

		// these should match each other 1 for 1 every step of the way
		// we can break out early if they ever don't
		if (watertransleafs != watersurfleafs) break;
	}

	if (watertransleafs == watersurfleafs)
	{
		Con_DPrintf ("Map %s has translucent water visibility!!!\n", mod->name);
		return true;
	}

NoTrans:;
	// no trans
	Con_DPrintf ("Map %s is not vis'ed for translucent water: transleafs: %i  surfleafs: %i\n", mod->name, watertransleafs, watersurfleafs);
	return false;
}


/*
====================
Mod_SurfMPFromVecs

Because we don;t have the poly data yet we need to calculate this one on the fly
====================
*/
void Mod_SurfMPFromVecs (model_t *mod, msurface_t *surf, vec3_t mp)
{
	float *vec;
	medge_t *r_pedge;
	int i;

	mp[0] = mp[1] = mp[2] = 0;

	for (i = 0; i < surf->numedges; i++)
	{
		if (mod->surfedges[surf->firstedge + i] > 0)
		{
			r_pedge = &mod->edges[mod->surfedges[surf->firstedge + i]];
			vec = mod->vertexes[r_pedge->v[0]].position;
		}
		else
		{
			r_pedge = &mod->edges[-mod->surfedges[surf->firstedge + i]];
			vec = mod->vertexes[r_pedge->v[1]].position;
		}

		mp[0] += vec[0] / (float) surf->numedges;
		mp[1] += vec[1] / (float) surf->numedges;
		mp[2] += vec[2] / (float) surf->numedges;
	}
}


void Mod_AddSurfacePointToPVS (model_t *mod, mleaf_t *leaf, vec3_t point, float zofs)
{
	int i;
	mleaf_t *newleaf;

	// chthon's hairy bum
	if (!leaf->visdata)
		return;

	// offset the z
	point[2] += zofs;

	// get the leaf
	newleaf = Mod_PointInLeaf (point, mod);

	// gollum's cock
	if (newleaf->contents == CONTENTS_SOLID)
		return;

	// sauron on a stick
	if (!newleaf->visdata)
		return;

	// accumulate the visibility
	// add 1 here cos we added 1 when setting up the decompressed vis to
	// cover a number of leafs that is not a multiple of 8
	for (i = 0; i < mod->numleafs + 1; i++)
		leaf->visdata[i] |= newleaf->visdata[i];
}


/*
====================
Mod_SetAWT

Calculates auto-water translucency for a model.

Goes through each empty leaf, checks the surfs in it's pvs, if it finds one with a liquid texture on it
it moves down 8 units, checks the contents, and if they're liquid accumulates the vis.  Then it moves down
120 more units, checks the contents again, and if they're empty accumulates the vis again (this last is for
handling stuff like the pool in the start map).

Finally, based on the assumption that if leaf x is visible from leaf y, then leaf y should also be visible
from leaf x (seems reasonable enough, although may not be true in all cases) it walks through the empty leafs
we've just created and sets them as visible from all liquid leafs that are visible from them.

Note that on a modern PC we can just set r_novis 1 and be damned with it, it's more than capable of handling
the fillrate.  However, what it doesn't give you is the proper underwater entities being sent from the server,
which this does.
====================
*/
void Mod_SetAWT (model_t *mod)
{
	int i;
	int c;
	mleaf_t *leaf;
	mleaf_t *vleaf;
	msurface_t **mark;
	int j;
	vec3_t mp;

	// already has translucent water!
	if (mod->transwater) return;

	for (i = 0, leaf = mod->leafs; i < mod->numleafs; i++, leaf++)
	{
		// only take empty leafs
		if (leaf->contents != CONTENTS_EMPTY)
			continue;

		// no visdata!!!
		if (!leaf->visdata)
			continue;

		for (j = 0, vleaf = mod->leafs; j < mod->numleafs; j++, vleaf++)
		{
			// check visible leafs
			if (leaf->visdata[j >> 3] & (1 << (j & 7)))
			{
				// check surfs
				mark = vleaf->firstmarksurface;
				c = vleaf->nummarksurfaces;

				// look for water
				if (c)
				{
					do
					{
						// have we a liquid texture?
						if ((*mark)->texinfo->texture->name[0] == '*')
						{
							// got a water surf in the visibility info
							// get it's mid-point
							Mod_SurfMPFromVecs (mod, (*mark), mp);

							// add 8 units down to leaf
							Mod_AddSurfacePointToPVS (mod, leaf, mp, -8);

							// add 120 units down to leaf
							Mod_AddSurfacePointToPVS (mod, leaf, mp, -120);

							// add 136 units up to vleaf
							Mod_AddSurfacePointToPVS (mod, vleaf, mp, -8);

							// add 120 units up to vleaf
							Mod_AddSurfacePointToPVS (mod, vleaf, mp, -120);
						}

						mark++;
					} while (--c);
				}
			}
		}
	}

	return;

	for (i = 0, leaf = mod->leafs; i < mod->numleafs; i++, leaf++)
	{
		// only take empty leafs
		if (leaf->contents != CONTENTS_EMPTY)
			continue;

		// no visdata!!!
		if (!leaf->visdata)
			continue;

		for (j = 0, vleaf = mod->leafs; j < mod->numleafs; j++, vleaf++)
		{
			// shub-niggurath's wiggly tentacle bits
			if (!vleaf->visdata) continue;

			// only interested in these content types here
			if (vleaf->contents == CONTENTS_WATER || vleaf->contents == CONTENTS_SLIME || vleaf->contents == CONTENTS_LAVA)
			{
				// if vleaf is in leaf's PVS, add leaf to vleaf's PVS
				if (leaf->visdata[j >> 3] & (1 << (j & 7)))
				{
					vleaf->visdata[j >> 3] |= (1 << (j & 7));
				}
			}
		}
	}
}


/*
=====================
IsWorldModel

Reliable detection of whether a model is the world.  If it doesn't have an "info_player_" entity
in it, it's not the world.
=====================
*/
qboolean IsWorldModel (lump_t *l)
{
	char *e = mod_base + l->fileofs;
	int i;

	if (!e) return false;

	for (i = 0; i < l->filelen - 11; i++, e++)
	{
		if (!(e[0] == 'i' || e[0] == 'I')) continue;
		if (!(e[1] == 'n' || e[1] == 'N')) continue;
		if (!(e[2] == 'f' || e[2] == 'F')) continue;
		if (!(e[3] == 'o' || e[3] == 'O')) continue;
		if (!(e[5] == 'p' || e[5] == 'P')) continue;
		if (!(e[6] == 'l' || e[6] == 'L')) continue;
		if (!(e[7] == 'a' || e[7] == 'A')) continue;
		if (!(e[8] == 'y' || e[8] == 'Y')) continue;
		if (!(e[9] == 'e' || e[9] == 'E')) continue;
		if (!(e[10] == 'r' || e[10] == 'R')) continue;

		if (e[4] != '_') continue;
		if (e[11] != '_') continue;

		// found an info_player_ entity
		return true;
	}

	// didn;t find anything
	return false;
}


/*
=================
Mod_LoadBrushModel
=================
*/
void Mod_LoadBrushModel (model_t *mod, void *buffer)
{
	int			i, j;
	dheader_t	*header;
	dmodel_t 	*bm;
	
	loadmodel->type = mod_brush;
	
	header = (dheader_t *)buffer;

	i = LittleLong (header->version);

	if (i != BSPVERSION) Sys_Error ("Mod_LoadBrushModel: %s has wrong version number (%i should be %i)", mod->name, i, BSPVERSION);

	// swap all the lumps
	mod_base = (byte *)header;

	for (i = 0; i < sizeof (dheader_t) / 4; i++)
		((int *) header)[i] = LittleLong (((int *) header)[i]);

	// load into heap
	Mod_LoadVertexes (&header->lumps[LUMP_VERTEXES]);
	Mod_LoadEdges (&header->lumps[LUMP_EDGES]);
	Mod_LoadSurfedges (&header->lumps[LUMP_SURFEDGES]);
	Mod_LoadTextures (&header->lumps[LUMP_TEXTURES]);
	Mod_LoadLighting (&header->lumps[LUMP_LIGHTING]);
	Mod_LoadPlanes (&header->lumps[LUMP_PLANES]);
	Mod_LoadTexinfo (&header->lumps[LUMP_TEXINFO]);
	Mod_LoadFaces (&header->lumps[LUMP_FACES]);
	Mod_LoadMarksurfaces (&header->lumps[LUMP_MARKSURFACES]);
	Mod_LoadVisibilityAndLeafs (&header->lumps[LUMP_LEAFS], &header->lumps[LUMP_VISIBILITY]);
	Mod_LoadNodes (&header->lumps[LUMP_NODES]);
	Mod_LoadClipnodes (&header->lumps[LUMP_CLIPNODES]);
	Mod_LoadEntities (&header->lumps[LUMP_ENTITIES]);
	Mod_LoadSubmodels (&header->lumps[LUMP_MODELS]);

	Mod_MakeHull0 ();

	if (IsWorldModel (&header->lumps[LUMP_ENTITIES]))
	{
		// check for translucent water
		mod->transwater = Mod_DetectTransWater (mod);

		// commented out because some mission pack authors have been known to use dual water layers with
		// weird visibility hacks to work around transition effects.  yup - fucking hipnotic again.
		// Mod_SetAWT (mod);
	}
	else mod->transwater = false;

	mod->numframes = 2;		// regular and alternate animation

	// set up the submodels (FIXME: this is confusing)
	for (i = 0; i < mod->numsubmodels; i++)
	{
		bm = &mod->submodels[i];

		mod->hulls[0].firstclipnode = bm->headnode[0];

		for (j = 1; j < MAX_MAP_HULLS; j++)
		{
			mod->hulls[j].firstclipnode = bm->headnode[j];
			mod->hulls[j].lastclipnode = mod->numclipnodes-1;
		}

		mod->firstmodelsurface = bm->firstface;
		mod->nummodelsurfaces = bm->numfaces;

		VectorCopy (bm->maxs, mod->maxs);
		VectorCopy (bm->mins, mod->mins);

		mod->radius = RadiusFromBounds (mod->mins, mod->maxs);

		mod->numleafs = bm->visleafs;

		if (i < mod->numsubmodels - 1)
		{
			// duplicate the basic information
			char	name[10];

			sprintf (name, "*%i", i + 1);
			loadmodel = Mod_FindName (name);
			*loadmodel = *mod;
			strcpy (loadmodel->name, name);
			mod = loadmodel;
		}
	}
}

/*
==============================================================================

ALIAS MODELS

==============================================================================
*/

aliashdr_t	*pheader;

stvert_t	stverts[MAXALIASVERTS];
mtriangle_t	triangles[MAXALIASTRIS];

// a pose is a single set of vertexes.  a frame may be
// an animating sequence of poses
trivertx_t	*poseverts[MAXALIASFRAMES];
int			posenum;

byte		**player_8bit_texels_tbl;
byte		*player_8bit_texels;

/*
=================
Mod_LoadAliasFrame
=================
*/
// alias bbox fix
int aliasbboxmins[3], aliasbboxmaxs[3];


void *Mod_LoadAliasFrame (void *pin, maliasframedesc_t *frame)
{
	trivertx_t		*pframe, *pinframe;
	int				i, j;
	daliasframe_t	*pdaliasframe;
	
	pdaliasframe = (daliasframe_t *)pin;

	strcpy (frame->name, pdaliasframe->name);
	frame->firstpose = posenum;
	frame->numposes = 1;

	for (i = 0; i < 3; i++)
	{
		// these are byte values, so we don't have to worry about
		// endianness
		frame->bboxmin.v[i] = pdaliasframe->bboxmin.v[i];
		frame->bboxmax.v[i] = pdaliasframe->bboxmax.v[i];	// bug - was min in the lvalue

		aliasbboxmins[i] = frame->bboxmin.v[i] > aliasbboxmins[i] ? aliasbboxmins[i] : frame->bboxmin.v[i];
		aliasbboxmaxs[i] = frame->bboxmax.v[i] < aliasbboxmaxs[i] ? aliasbboxmaxs[i] : frame->bboxmax.v[i];
	}

	pinframe = (trivertx_t *) (pdaliasframe + 1);

	poseverts[posenum] = pinframe;
	posenum++;

	pinframe += pheader->numverts;

	return (void *)pinframe;
}


/*
=================
Mod_LoadAliasGroup
=================
*/
void *Mod_LoadAliasGroup (void * pin,  maliasframedesc_t *frame)
{
	daliasgroup_t		*pingroup;
	int					i, numframes;
	daliasinterval_t	*pin_intervals;
	void				*ptemp;
	
	pingroup = (daliasgroup_t *)pin;

	numframes = LittleLong (pingroup->numframes);

	frame->firstpose = posenum;
	frame->numposes = numframes;

	for (i = 0; i < 3; i++)
	{
		// these are byte values, so we don't have to worry about endianness
		frame->bboxmin.v[i] = pingroup->bboxmin.v[i];
		frame->bboxmax.v[i] = pingroup->bboxmax.v[i];

		aliasbboxmins[i] = frame->bboxmin.v[i] > aliasbboxmins[i] ? aliasbboxmins[i] : frame->bboxmin.v[i];
		aliasbboxmaxs[i] = frame->bboxmax.v[i] < aliasbboxmaxs[i] ? aliasbboxmaxs[i] : frame->bboxmax.v[i];
	}

	pin_intervals = (daliasinterval_t *)(pingroup + 1);

	frame->interval = LittleFloat (pin_intervals->interval);

	pin_intervals += numframes;

	ptemp = (void *)pin_intervals;

	for (i=0 ; i<numframes ; i++)
	{
		poseverts[posenum] = (trivertx_t *)((daliasframe_t *)ptemp + 1);
		posenum++;

		ptemp = (trivertx_t *)((daliasframe_t *)ptemp + 1) + pheader->numverts;
	}

	return ptemp;
}

//=========================================================

/*
===============
Mod_LoadAllSkins
===============
*/
void *Mod_LoadAllSkins (int numskins, daliasskintype_t *pskintype)
{
	int		i, j, k;
	char	name[32];
	int		s;
	byte	*copy;
	byte	*skin;
	daliasskingroup_t		*pinskingroup;
	int		groupskins;
	daliasskininterval_t	*pinskinintervals;
	unsigned int skinflags;
	
	skin = (byte *) (pskintype + 1);

	// prevent code stretching over to a third line!!!
	skinflags = (TEX_MIPMAP | TEX_ALIAS | TEX_LUMASPEC);

	// identify viewmodel skins
	if ((loadmodel->name[6] == 'v' || loadmodel->name[6] == 'V') && loadmodel->name[7] == '_')
		skinflags |= TEX_VIEWMODEL;

	if (numskins < 1 || numskins > MAX_SKINS)
		Sys_Error ("Mod_LoadAliasModel: Invalid # of skins: %d\n", numskins);

	s = pheader->skinwidth * pheader->skinheight;

	for (i = 0; i < numskins; i++)
	{
		if (pskintype->type == ALIAS_SKIN_SINGLE)
		{
			sprintf (name, "%s_%i", loadmodel->name, i);

			pheader->glt[i][0] = pheader->glt[i][1] = pheader->glt[i][2] = pheader->glt[i][3] =
				GL_LoadTexture (name, pheader->skinwidth, pheader->skinheight, (byte *) (pskintype + 1), skinflags);

			pskintype = (daliasskintype_t *) ((byte *) (pskintype + 1) + s);
		}
		else
		{
			// animating skin group.  yuck.
			pskintype++;
			pinskingroup = (daliasskingroup_t *) pskintype;
			groupskins = LittleLong (pinskingroup->numskins);
			pinskinintervals = (daliasskininterval_t *) (pinskingroup + 1);

			pskintype = (void *) (pinskinintervals + groupskins);

			for (j = 0; j < groupskins; j++)
			{
				sprintf (name, "%s_%i_%i", loadmodel->name, i, j);

				pheader->glt[i][j & 3] = GL_LoadTexture (name, pheader->skinwidth, 
					pheader->skinheight, (byte *) (pskintype), skinflags);

				pskintype = (daliasskintype_t *) ((byte *) (pskintype) + s);
			}

			// fill up to 4 skins
			// hurrr hurrr.  he said "4 skins".
			for (k = j; j < 4; j++) pheader->glt[i][j & 3] = pheader->glt[i][j - k]; 
		}
	}

	return (void *) pskintype;
}

//=========================================================================

/*
=================
Mod_LoadAliasModel
=================
*/
void Mod_LoadAliasModel (model_t *mod, void *buffer)
{
	int					i, j;
	mdl_t				*pinmodel;
	stvert_t			*pinstverts;
	dtriangle_t			*pintriangles;
	int					version, numframes, numskins;
	int					size;
	daliasframetype_t	*pframetype;
	daliasskintype_t	*pskintype;
	int					start, end, total;
	char				*tempname;
	
	start = Hunk_LowMark ();

	pinmodel = (mdl_t *)buffer;

	version = LittleLong (pinmodel->version);

	if (version != ALIAS_VERSION) Sys_Error ("%s has wrong version number (%i should be %i)", mod->name, version, ALIAS_VERSION);

	// allocate space for a working header, plus all the data except the frames,
	// skin and group info
	size = 	sizeof (aliashdr_t) + (LittleLong (pinmodel->numframes) - 1) * sizeof (pheader->frames[0]);
	pheader = Hunk_AllocName (size, loadname);
	
	mod->flags = LittleLong (pinmodel->flags);

	// endian-adjust and copy the data, starting with the alias model header
	pheader->boundingradius = LittleFloat (pinmodel->boundingradius);
	pheader->numskins = LittleLong (pinmodel->numskins);
	pheader->skinwidth = LittleLong (pinmodel->skinwidth);
	pheader->skinheight = LittleLong (pinmodel->skinheight);

	if (pheader->skinheight > MAX_LBM_HEIGHT) Sys_Error ("model %s has a skin taller than %d", mod->name, MAX_LBM_HEIGHT);

	pheader->numverts = LittleLong (pinmodel->numverts);

	if (pheader->numverts <= 0) Sys_Error ("model %s has no vertices", mod->name);
	if (pheader->numverts > MAXALIASVERTS) Sys_Error ("model %s has too many vertices", mod->name);

	pheader->numtris = LittleLong (pinmodel->numtris);

	if (pheader->numtris <= 0) Sys_Error ("model %s has no triangles", mod->name);

	pheader->numframes = LittleLong (pinmodel->numframes);
	numframes = pheader->numframes;
	if (numframes < 1) Sys_Error ("Mod_LoadAliasModel: Invalid # of frames: %d\n", numframes);

	pheader->size = LittleFloat (pinmodel->size) * ALIAS_BASE_SIZE_RATIO;
	mod->synctype = LittleLong (pinmodel->synctype);
	mod->numframes = pheader->numframes;

	for (i = 0; i < 3; i++)
	{
		pheader->scale[i] = LittleFloat (pinmodel->scale[i]);
		pheader->scale_origin[i] = LittleFloat (pinmodel->scale_origin[i]);
		pheader->eyeposition[i] = LittleFloat (pinmodel->eyeposition[i]);
	}

	// load the skins
	pskintype = (daliasskintype_t *)&pinmodel[1];
	pskintype = Mod_LoadAllSkins (pheader->numskins, pskintype);

	// load base s and t vertices
	pinstverts = (stvert_t *)pskintype;

	for (i = 0; i < pheader->numverts; i++)
	{
		stverts[i].onseam = LittleLong (pinstverts[i].onseam);
		stverts[i].s = LittleLong (pinstverts[i].s);
		stverts[i].t = LittleLong (pinstverts[i].t);
	}

	// load triangle lists
	pintriangles = (dtriangle_t *) &pinstverts[pheader->numverts];

	for (i = 0; i < pheader->numtris; i++)
	{
		triangles[i].facesfront = LittleLong (pintriangles[i].facesfront);

		for (j = 0; j < 3; j++) triangles[i].vertindex[j] = LittleLong (pintriangles[i].vertindex[j]);
	}

	// load the frames
	posenum = 0;
	pframetype = (daliasframetype_t *)&pintriangles[pheader->numtris];

	// initial bbox
	aliasbboxmins[0] = aliasbboxmins[1] = aliasbboxmins[2] = 99999;
	aliasbboxmaxs[0] = aliasbboxmaxs[1] = aliasbboxmaxs[2] = -99999;

	for (i = 0; i < numframes; i++)
	{
		aliasframetype_t frametype = LittleLong (pframetype->type);

		if (frametype == ALIAS_SINGLE)
			pframetype = (daliasframetype_t *) Mod_LoadAliasFrame (pframetype + 1, &pheader->frames[i]);
		else pframetype = (daliasframetype_t *) Mod_LoadAliasGroup (pframetype + 1, &pheader->frames[i]);
	}

	pheader->numposes = posenum;
	mod->type = mod_alias;

	// FIXME: do this right
	// MH - consider it done.
	for (i = 0; i < 3; i++)
	{
		mod->mins[i] = aliasbboxmins[i] * pheader->scale[i] + pheader->scale_origin[i];
		mod->maxs[i] = aliasbboxmaxs[i] * pheader->scale[i] + pheader->scale_origin[i];
	}

	// build the draw lists
	GL_MakeAliasModelDisplayLists (mod, pheader);

	// set effects
	// initially none
	mod->effects = 0;

	// rip out the "progs" directory so that we can more conveniently compare
	tempname = &mod->name[6];

	// OK, this is a little hacky as it uses hard-coded model names in the engine, which restricts customisation
	// outside of the engine.  for completeness, i've added in rogue and hipnotic as well as id1
	if (!stricmp (tempname, "flame.mdl")) mod->effects |= (EF_TORCHFLARE | EF_FLICKER);
	if (!stricmp (tempname, "flame2.mdl")) mod->effects |= (EF_FIREFLARE | EF_FLICKER);
	if (!stricmp (tempname, "candle.mdl")) mod->effects |= (EF_TORCHFLARE | EF_FLICKER);
	if (!stricmp (tempname, "lspike.mdl")) mod->effects |= EF_FIREFLARE;
	if (!stricmp (tempname, "missile.mdl")) mod->effects |= (EF_ROCKETFLARE | EF_FLICKER);

	// move the complete, relocatable alias model to the cache
	end = Hunk_LowMark ();
	total = end - start;
	
	Cache_Alloc (&mod->cache, total, loadname);

	if (!mod->cache.data) return;

	Q_memcpy (mod->cache.data, pheader, total);

	Hunk_FreeToLowMark (start);
}

//=============================================================================

/*
=================
Mod_LoadSpriteFrame
=================
*/
void * Mod_LoadSpriteFrame (void * pin, mspriteframe_t **ppframe, int framenum)
{
	dspriteframe_t		*pinframe;
	mspriteframe_t		*pspriteframe;
	int					i, width, height, size, origin[2];
	unsigned short		*ppixout;
	byte				*ppixin;
	char				name[64];

	pinframe = (dspriteframe_t *)pin;

	width = LittleLong (pinframe->width);
	height = LittleLong (pinframe->height);
	size = width * height;

	pspriteframe = Hunk_AllocName (sizeof (mspriteframe_t),loadname);

	Q_memset (pspriteframe, 0, sizeof (mspriteframe_t));

	*ppframe = pspriteframe;

	pspriteframe->width = width;
	pspriteframe->height = height;
	origin[0] = LittleLong (pinframe->origin[0]);
	origin[1] = LittleLong (pinframe->origin[1]);

	pspriteframe->up = origin[1];
	pspriteframe->down = origin[1] - height;
	pspriteframe->left = origin[0];
	pspriteframe->right = width + origin[0];

	sprintf (name, "%s_%i", loadmodel->name, framenum);
	pspriteframe->glt = GL_LoadTexture (name, width, height, (byte *)(pinframe + 1), (TEX_MIPMAP | TEX_SPRITE));

	return (void *)((byte *)pinframe + sizeof (dspriteframe_t) + size);
}


/*
=================
Mod_LoadSpriteGroup
=================
*/
void * Mod_LoadSpriteGroup (void * pin, mspriteframe_t **ppframe, int framenum)
{
	dspritegroup_t		*pingroup;
	mspritegroup_t		*pspritegroup;
	int					i, numframes;
	dspriteinterval_t	*pin_intervals;
	float				*poutintervals;
	void				*ptemp;

	pingroup = (dspritegroup_t *)pin;

	numframes = LittleLong (pingroup->numframes);

	pspritegroup = Hunk_AllocName (sizeof (mspritegroup_t) +
				(numframes - 1) * sizeof (pspritegroup->frames[0]), loadname);

	pspritegroup->numframes = numframes;

	*ppframe = (mspriteframe_t *)pspritegroup;

	pin_intervals = (dspriteinterval_t *)(pingroup + 1);

	poutintervals = Hunk_AllocName (numframes * sizeof (float), loadname);

	pspritegroup->intervals = poutintervals;

	for (i=0 ; i<numframes ; i++)
	{
		*poutintervals = LittleFloat (pin_intervals->interval);
		if (*poutintervals <= 0.0)
			Sys_Error ("Mod_LoadSpriteGroup: interval<=0");

		poutintervals++;
		pin_intervals++;
	}

	ptemp = (void *)pin_intervals;

	for (i=0 ; i<numframes ; i++)
	{
		ptemp = Mod_LoadSpriteFrame (ptemp, &pspritegroup->frames[i], framenum * 100 + i);
	}

	return ptemp;
}


/*
=================
Mod_LoadSpriteModel
=================
*/
void Mod_LoadSpriteModel (model_t *mod, void *buffer)
{
	int					i;
	int					version;
	dsprite_t			*pin;
	msprite_t			*psprite;
	int					numframes;
	int					size;
	dspriteframetype_t	*pframetype;
	char				*tempname;

	pin = (dsprite_t *)buffer;

	version = LittleLong (pin->version);
	if (version != SPRITE_VERSION)
		Sys_Error ("%s has wrong version number "
				 "(%i should be %i)", mod->name, version, SPRITE_VERSION);

	numframes = LittleLong (pin->numframes);

	size = sizeof (msprite_t) +	(numframes - 1) * sizeof (psprite->frames);

	psprite = Hunk_AllocName (size, loadname);

	mod->cache.data = psprite;

	psprite->type = LittleLong (pin->type);
	psprite->maxwidth = LittleLong (pin->width);
	psprite->maxheight = LittleLong (pin->height);
	psprite->beamlength = LittleFloat (pin->beamlength);
	mod->synctype = LittleLong (pin->synctype);
	psprite->numframes = numframes;

	mod->mins[0] = mod->mins[1] = -psprite->maxwidth/2;
	mod->maxs[0] = mod->maxs[1] = psprite->maxwidth/2;
	mod->mins[2] = -psprite->maxheight/2;
	mod->maxs[2] = psprite->maxheight/2;
	
//
// load the frames
//
	if (numframes < 1)
		Sys_Error ("Mod_LoadSpriteModel: Invalid # of frames: %d\n", numframes);

	mod->numframes = numframes;

	pframetype = (dspriteframetype_t *)(pin + 1);

	for (i=0 ; i<numframes ; i++)
	{
		spriteframetype_t	frametype;

		frametype = LittleLong (pframetype->type);
		psprite->frames[i].type = frametype;

		if (frametype == SPR_SINGLE)
		{
			pframetype = (dspriteframetype_t *)
					Mod_LoadSpriteFrame (pframetype + 1,
										 &psprite->frames[i].frameptr, i);
		}
		else
		{
			pframetype = (dspriteframetype_t *)
					Mod_LoadSpriteGroup (pframetype + 1,
										 &psprite->frames[i].frameptr, i);
		}
	}

	mod->type = mod_sprite;

	tempname = &mod->name[6];

	// set effects
	mod->effects = 0;

	if (!stricmp (tempname, "s_light.spr")) mod->effects |= EF_S_LIGHT;
	if (!stricmp (tempname, "s_bubble.spr")) mod->effects |= EF_BUBBLESPRITE;
}

//=============================================================================

/*
================
Mod_Print
================
*/
void Mod_Print (void)
{
	int		i;
	model_t	*mod;

	Con_Printf ("Cached models:\n");
	for (i=0, mod=mod_known ; i < mod_numknown ; i++, mod++)
	{
		Con_Printf ("%8p : %s\n",mod->cache.data, mod->name);
	}
}


